

from .utils import logger_config,DirCheck,RunCMD


def SNVCalling_process(sample,bam,fasta,outvcf,pon,germline,config,chrom = None):
    logger = logger_config("SNVCalling")
    mutect2_cmd = config.gatk_cmd("Mutect2")

    for single_sample,single_bam in zip(sample,bam):
        logger.debug(f"Match BAM Path: [{single_sample}]-{single_bam}")
        mutect2_cmd.extend(
            [
                "-I",
                single_bam
            ]
        )
    raw_vcf = outvcf[:-7] + '.raw.vcf.gz'
    mutect2_cmd.extend(
        [
            "-R",
            fasta,
            "--minimum-mapping-quality",
            10,
            "-tumor",
            "tumor",
            "--native-pair-hmm-threads",
            "10",
            "--output",
            raw_vcf,
            "--germline-resource",
            f"{germline}",
            "--panel-of-normals",
            f"{pon}",
        ]
    )
    if chrom != None:
        mutect2_cmd.extend(
            [
                "-L",
                f"{chrom}",]
        )
    sample_str = ','.join(sample)
    RunCMD(mutect2_cmd,"Mutect2","",logger)

    filter_cmd = config.gatk_cmd("FilterMutectCalls")
    filter_cmd.extend(
        [
            "-V",
            raw_vcf,
            "-R",
            fasta,
            "-O",
            outvcf,
            "--min-median-base-quality",
            10
        ]
    )
    RunCMD(filter_cmd,"FilterMutectCalls","",logger)


import logging

def logger_config(logging_name):
    
    logger = logging.getLogger(logging_name)
    logger.setLevel(level=logging.DEBUG)
    console = logging.StreamHandler()
    console.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    console.setFormatter(formatter)
    logger.addHandler(console)
    return logger


import pysam
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any
from subprocess import run,PIPE
import subprocess


from .log import logger_config

def BulidConfig(samtools = None,gatk = None,picard = None,threads = 1):
    logger = logger_config("CheckSoftware")
    if samtools != None:
        samtools = check_software('samtools',samtools,logger = logger)
    if picard != None:
        picard = check_software('picard',picard,logger = logger)
    gatk = check_software('gatk',gatk,logger = logger)
    config = Config.set_path(picard = picard, gatk = gatk, samtools = samtools, threads = threads )
    return config

def check_software(software_name, software_path,logger = None):
    if not software_path:
        try:
            output = subprocess.check_output(["which", f"{software_name}"], stderr=subprocess.STDOUT)
            output = output.decode().rstrip('\n')
            if logger != None:
                logger.info(f"Check {software_name} : Find in {output}")
            return output
        except subprocess.CalledProcessError as e:
            if logger != None:
                logger.debug(f"{software_name} may not be installed, please use install {software_name}\n Or set --{software_name} PATH_TO_{software_name}")
    else:
        if os.path.exists(software_path):
            if logger != None:
                logger.info(f"Check {software_name} : Find in {software_path}")
            return software_path
        else:
            raise FileNotFoundError(f"{software_name} not exist")

@dataclass
class Config:
    picard: Path
    gatk: Path
    samtools: Path
    threads: int
    def set_path(picard,gatk,samtools,threads):
        return Config(picard = picard,
                     gatk = gatk,
                     samtools = samtools,
                     threads = threads)
    
    def picard_cmd(self, command: str, tmp_dir: Path, mem: str = "100g"):
        return [
            "java",
            f"-Djava.io.tmp_dir={tmp_dir}",
            f"-Xmx{mem}",
            "-XX:+UseParallelGC",
            "-XX:GCTimeLimit=20",
            "-XX:GCHeapFreeLimit=10",
            "-jar",
            self.picard,
            command,
            "--TMP_DIR",
            tmp_dir,
            "--VALIDATION_STRINGENCY",
            "SILENT",
            "--VERBOSITY",
            "WARNING",
            "--QUIET",
            "true",
        ]
    def gatk_cmd(self,command: str, tmp_dir: str = None,mem: str = "100g"):
        if tmp_dir != None:
            return [
                self.gatk,
                "--java-options",
                f"-Xmx{mem}",
                command,
                "--tmp-dir",
                tmp_dir,
            ]
        else:
            return [
                self.gatk,
                "--java-options",
                f"-Xmx{mem}",
                command,
            ]
    def samtools_cmd(self, command: str):
        return [
            self.samtools,
            command,
            "-@",
            self.threads,
        ]

def RunCMD(cmd: list[Any],name: str,id: str,logger):
    cmd = [str(arg) for arg in cmd]
    proc = run(cmd, capture_output=True, text=True)
    if proc.returncode != 0:
        raise Exception(f"Error running {name}:\n\t{proc.stderr}")
    else:
        logger.info(f"{id}-{name}: completed")

def check_bam_index(bam):
    bai_file = f"{bam}.bai"
    logger = logger_config("BAMIndexCheck")
    if not os.path.exists(bai_file):
        raise Exception(f"{bam} without index")
    logger.info(f"Find BAM Index: {bai_file}")
    

def LoadChromFromBAM(bam, only_autosome = True):
    logger = logger_config("LoadChromFromBAM")
    check_bam_index(bam)
    inputfile = pysam.AlignmentFile(bam,"rb")
    chrom_tuple = inputfile.references
    inputfile.close()
    if only_autosome:
        chrom_tuple_tmp = [chrom for chrom in chrom_tuple if chrom.startswith('chr') and 'Un' not in chrom ]
        if len(chrom_tuple_tmp) == 0 :
            chrom_tuple_tmp = [chrom for chrom in chrom_tuple if "K" not in chrom and "G" not in chrom and "L" not in chrom and "J" not in chrom and "V" not in chrom and "Q" not in chrom and "N" not in chrom and "h" not in chrom and '.' not in chrom and 'Z' not in chrom]
        chrom_str = "[" + ','.join(chrom_tuple_tmp) + "]"
        logger.info(f"Find CHROM: {chrom_str}")
        return chrom_tuple_tmp
    else:
        chrom_str = "[" + ','.join(chrom_tuple) + "]"
        logger.info(f"Find CHROM: {chrom_str}")
        return chrom_tuple

def DirCheck(path, create=False):
    if os.path.isdir(path):
        return True
    else:
        if create:
            try:
                os.makedirs(path)
                return True
            except OSError:
                raise OSError(f"Failed to create {path}")
        return False


import pysam
import os


from multiprocess import Pool

from .utils import LoadChromFromBAM,logger_config,DirCheck


def SplitChromBAM_process(bam,only_autosome,sampleid,out_dir,threads = 1):#chrom,bam,output_dir,sample_name):
    DirCheck(out_dir,create=True)
    chrom_list = LoadChromFromBAM(bam,only_autosome)
    logger = logger_config(f'SplitChromBAM')
    if threads == 1:
        for chrom in chrom_list:
            split_bam(bam,chrom,sampleid,out_dir,logger)
    else:
        split_pool = Pool(threads)
        for chrom in chrom_list:
            split_pool.apply_async(split_bam, args=(bam,chrom,sampleid,out_dir,logger))
        split_pool.close()
        split_pool.join()


def split_bam(bam,chrom,sample_name,output_dir,logger):
    inputfile = pysam.AlignmentFile(bam, "rb")
    tmp_bam = os.path.join(output_dir, '%s.%s.bam'%(sample_name,chrom))
    output = pysam.AlignmentFile(tmp_bam, 'wb', template = inputfile)
    m = 0
    for read in inputfile.fetch(str(chrom), multiple_iterators=True):
        output.write(read)
        m+=1
        if m>1000000:
            break
    logger.info(f'{chrom} finish')
    inputfile.close()
    output.close()



#!/usr/bin/env python3
import click
from .SplitChromBAM import SplitChromBAM_process
from .PerpareBAMforCalling import PerpareBAMforCalling_process
from .SNVCalling import SNVCalling_process
from .CallBack import CallBack_process,BulidResult
from .utils import BulidConfig


@click.group()
def main():
    pass

@main.command(name="SplitChromBAM",no_args_is_help=True)
@click.option("-b","--bam", type = click.Path(exists=True, file_okay=True, dir_okay=False), required = True, help = "")
@click.option("-s","--sample", type = str, required = True, help = "Sample ID ")
@click.option("-o","--outdir", type = str, required = True,  help = "Output dir")
@click.option("-@","--threads", type = int, default = 10, help = "Sets the number of threads")
@click.option('--only_autosome', is_flag=True,default=False,help="Only analysis autosome")
def SplitChromBAM(
    bam: str,
    sample: str,
    outdir: str,
    threads: int = 10,
    only_autosome: bool = False,
):
    SplitChromBAM_process(bam,only_autosome,sample,outdir,threads = threads)

@main.command(name="PerpareBAMforCalling",no_args_is_help=True)
@click.option("-b","--bam", type = click.Path(exists=True, file_okay=True, dir_okay=False), required = True, help = "")
@click.option("-o","--outdir", type = str, required = True,  help = "Output dir")
@click.option("-s","--proxy", type = str, required = True, help = "Sample ID ")
@click.option("-f","--fasta", type = click.Path(exists=True, file_okay=True, dir_okay=False), required = True, help = "Refence fasta")
@click.option("-d","--dbsnp", type = click.Path(exists=True, file_okay=True, dir_okay=False), required = True, help = "dbsnp")
@click.option("-c","--barcode", type = str, default = 'CR', help = "Cell Barcode in Bam file (ex. CR for 10X Genomics)")
@click.option("-u","--umi", type = str, default ='UR', help = "UMI(Molecular Barcodes) in Bam file (ex. UR for 10X Genomics)")
@click.option('--stereo', is_flag=True,default=False, help="Please check your data is stereo(barcode is Cx and Cy) or not")
@click.option('--gem',type = str, help="gem match your raw stereo bam")
@click.option("-x",'--xsetoff',type = int, help="gem_x + x_offset = bam_x")
@click.option("-y",'--ysetoff',type = int, help="gem_y + y_offset = bam_y")
@click.option('--tmpdir',type=str,help="Specify a temporary directory")
@click.option("-@","--threads", type = int, default = 10, help = "Sets the number of threads")
@click.option('--samtools',type=str,help="Specify samtools path")
@click.option('--picard',type=str,help="Specify picard.jar path")
@click.option('--gatk',type=str,help="Specify GATK path")
@click.option('--removetmp', is_flag=True,default=False, help="Remove all tmp file")
def PerpareBAMforCalling(
    bam: str,
    outdir: str,
    proxy: str,
    fasta: str,
    dbsnp: str,
    barcode: str = 'CB',
    umi: str = 'UB',
    stereo: bool = True,
    gem: str = None,
    xsetoff: int = 0,
    ysetoff: int = 0,
    tmpdir: str = None,
    samtools: str = None,
    gatk: str = None,
    picard: str = None,
    threads: int = 10,
    removetmp: bool = False,
    ):
    config = BulidConfig(samtools = samtools, gatk = gatk, picard = picard, threads= threads)
    PerpareBAMforCalling_process(bam,outdir,proxy,barcode,umi,stereo,gem,xsetoff,ysetoff,config,fasta,dbsnp,tmpdir,removetmp)


@main.command(name="SNVCalling",no_args_is_help=True)
@click.option("-s","--sample", multiple=True,required=True, help="Bam file with inde(ex: -b example.bam1 -b example.bam2)")
@click.option("-b","--bam", type=click.Path(exists=True, file_okay=True, dir_okay=False), multiple=True,required=True, help="Bam file with inde(ex: -b example.bam1 -b example.bam2)")
@click.option("-o","--outvcf", type = str, required = True,  help = "Output VCF")
@click.option("-f","--fasta", type = click.Path(exists=True, file_okay=True, dir_okay=False), required = True, help = "Refence fasta")
@click.option("--pon",type=click.Path(exists=True, file_okay=True, dir_okay=False),required = True,help="PON")
@click.option("--germline",type=click.Path(exists=True, file_okay=True, dir_okay=False),required = True, help="Germline source")
@click.option('--gatk',type=str,help="Specify GATK path")
@click.option("-L","--chrom",type = str, required = False,  help = "Specific CHROM")
def SNVCalling(
    sample: str,
    bam: str,
    fasta: str,
    outvcf: str,
    pon: str,
    germline: str,
    gatk: str,
    chrom: str = None,
    ):
    if len(sample) != len(bam):
        raise Exception(f"The lenth of BAM not match with SAMPLE")
    config = BulidConfig(gatk = gatk,threads= 1)
    SNVCalling_process(sample,bam,fasta,outvcf,pon,germline,config,chrom)



@main.command(name="CallBack",no_args_is_help=True)
@click.option("-b","--bam", type=click.Path(exists=True, file_okay=True, dir_okay=False), required=True, help="Bam file with inde(ex: -b example.bam1 -b example.bam2)")
@click.option("-v","--vcf", type=click.Path(exists=True, file_okay=True, dir_okay=False), required = True,  help = "Output VCF")
@click.option("-s","--sample",required=True, help="Bam file with inde(ex: -b example.bam1 -b example.bam2)")
@click.option("-o","--outdir", type = str, required = True,  help = "Output dir")
@click.option('--tmpdir',type=str,help="Specify a temporary directory")
@click.option('--only_autosome', is_flag=True,default=False,help="Only analysis autosome")
@click.option('--stereo', is_flag=True,default=False, help="Please check your data is stereo(barcode is Cx and Cy) or not")
@click.option("-c","--barcode", type = str, default = 'CB', help = "Cell Barcode in Bam file (ex. CR for 10X Genomics)")
@click.option("-u","--umi", type = str, default ='UB', help = "UMI(Molecular Barcodes) in Bam file (ex. UR for 10X Genomics)")
@click.option("-@","--threads", type = int, default = 10, help = "Sets the number of threads")
@click.option("--qmap", type = int, default = 10, help = "Sets the number of qmap")
@click.option("-x",'--xsetoff',default = None,type = int, help="gem_x + x_offset = bam_x")
@click.option("-y",'--ysetoff',default = None,type = int, help="gem_y + y_offset = bam_y")
@click.option('--binsize',type = int,default = None, help="binsize")
@click.option('--removetmp', is_flag=True,default=False, help="Remove all tmp file")
def CallBack(
    bam: str,
    vcf: str,
    sample: str,
    outdir: str,
    stereo: bool = True,
    only_autosome: bool = True,
    barcode: str = 'CB',
    umi: str = 'UB',
    tmpdir: str = None,
    qmap: int = 10,
    threads: int = 10,
    xsetoff: int = None,
    ysetoff: int = None,
    binsize : int = None,
    removetmp: bool = False,
):
    CallBack_process(bam,only_autosome,sample,outdir,tmpdir,vcf,barcode,umi,qmap,stereo,threads = threads)
    BulidResult(sample = sample ,sample_tmp_path = tmpdir,sample_matrix_path = outdir,x_offset = xsetoff,y_offset = ysetoff,bin_size = binsize,removetmp = removetmp)


import pysam 
from pysam import VariantFile
from multiprocess import Pool
import pandas as pd 
import os
import glob
import shutil
from .utils import LoadChromFromBAM,logger_config,DirCheck

def fillter(read):
    m = 1
    if 'weak_evidence' in read.filter or 'germline' in read.filter or 'strand_bias' in read.filter or 'slippage' in read.filter or 'contamination' in read.filter:
        m = 0
    return m


def CallBack_process(bam,only_autosome,sample,out_dir,tmp_dir,vcf,barcode_tag,umi_tag,qmap,stereo,threads = 1):
    DirCheck(out_dir,create=True)
    DirCheck(tmp_dir,create=True)

    chrom_list = LoadChromFromBAM(bam,only_autosome)
    logger = logger_config(f'CallBack')

    if stereo:
        callback = StereoBack
        logger.debug(f'Call back on stereo-seq')
    else:
        callback = DropseqBack
        logger.debug(f'Call back on drop-seq like')

    logger.info(f'Save stat in {tmp_dir}')

    if threads == 1:
        for chrom in chrom_list:
            callback(bam,vcf,sample,tmp_dir,chrom,barcode_tag,umi_tag,qmap,logger)
    else:
        split_pool = Pool(threads)
        for chrom in chrom_list:
            split_pool.apply_async(callback, args=(bam,vcf,sample,tmp_dir,chrom,barcode_tag,umi_tag,qmap,logger))
        split_pool.close()
        split_pool.join()


def StereoBack(bam,vcf,sample,outdir,chrom,barcode_tag,umi_tag,qmap,logger):
    logger.info(f'Start {chrom} calling...')
    inbam = pysam.AlignmentFile(bam,"rb")
    vcf_input = VariantFile(vcf,"r")
    alt_result_name = os.path.join(outdir,f'{sample}.{chrom}.alt.stat')
    ref_result_name = os.path.join(outdir,f'{sample}.{chrom}.ref.stat')
    alt_result = open(alt_result_name,"w")
    ref_result = open(ref_result_name,"w")
    total = 0
    for read in vcf_input.fetch(chrom,reopen=True):
        total += 1
        if all(len(allele) == 1 for allele in read.alleles) and fillter(read):
            vcf_pos = read.pos
            for i in inbam.pileup(chrom,vcf_pos-1,vcf_pos,truncate=True,min_mapping_quality=qmap):
                for pileupread in i.pileups:
                    if not pileupread.is_del and not pileupread.is_refskip and not pileupread.alignment.is_qcfail and not pileupread.alignment.is_duplicate:
                        read_base = pileupread.alignment.query_sequence[pileupread.query_position]
                        barcode = str(pileupread.alignment.get_tag('Cx'))+'_'+str(pileupread.alignment.get_tag('Cy'))
                        umi = str(pileupread.alignment.get_tag(umi_tag))
                        if read_base in read.alts:
                            snv_name = "%s_%s:%s>%s"%(chrom,vcf_pos,read.ref,read_base)
                            alt_result.write("%s,%s,%s\n"%(snv_name,barcode,umi))
                        if read_base == read.ref:
                            snv_name = "%s_%s:%s"%(chrom,vcf_pos,read.ref)
                            ref_result.write("%s,%s,%s\n"%(snv_name,barcode,umi))


    if total == 0:
        logger.warning(f'Not Found Any SNV in {chrom}')
    else:
        logger.debug(f'>> Finish {chrom} calling ::: Process {total} SNVs')

def DropseqBack(bam,vcf,sample,outdir,chrom,barcode_tag,umi_tag,qmap,logger):
    logger.info(f'Start {chrom} calling...')
    inbam = pysam.AlignmentFile(bam,"rb")
    vcf_input = VariantFile(vcf,"r")
    alt_result_name = os.path.join(outdir,f'{sample}.{chrom}.alt.stat')
    ref_result_name = os.path.join(outdir,f'{sample}.{chrom}.ref.stat')
    alt_result = open(alt_result_name,"w")
    ref_result = open(ref_result_name,"w")
    total = 0
    for read in vcf_input.fetch(chrom,reopen=True):
        total += 1
        if all(len(allele) == 1 for allele in read.alleles) and fillter(read):
            vcf_pos = read.pos
            for i in inbam.pileup(chrom,vcf_pos-1,vcf_pos,truncate=True,min_mapping_quality=qmap):
                for pileupread in i.pileups:
                    if not pileupread.is_del and not pileupread.is_refskip and not pileupread.alignment.is_qcfail and not pileupread.alignment.is_duplicate and pileupread.alignment.has_tag(barcode_tag) and pileupread.alignment.has_tag(umi_tag):
                        read_base = pileupread.alignment.query_sequence[pileupread.query_position]
                        barcode = str(pileupread.alignment.get_tag(barcode_tag))
                        umi = str(pileupread.alignment.get_tag(umi_tag))
                        if read_base in read.alts:
                            snv_name = "%s_%s:%s>%s"%(chrom,vcf_pos,read.ref,read_base)
                            alt_result.write("%s,%s,%s\n"%(snv_name,barcode,umi))
                        if read_base == read.ref:
                            snv_name = "%s_%s:%s"%(chrom,vcf_pos,read.ref)
                            ref_result.write("%s,%s,%s\n"%(snv_name,barcode,umi))

    if total == 0:
        logger.warning(f'Not Found Any SNV in {chrom}')
    else:
        logger.debug(f'>> Finish {chrom} calling ::: Process {total} SNVs')





def BulidResult(sample,sample_tmp_path,sample_matrix_path,x_offset,y_offset,bin_size,removetmp):
    # alt and supported reads
    alt_vcf_list = glob.glob(f"{sample_tmp_path}/*.alt.stat")
    df = pd.DataFrame(columns=['snv','barcode','umi'])
    for i in alt_vcf_list:
        if os.path.getsize(i):
            tmp=pd.read_csv(i,header=None)
            tmp.columns=['snv','barcode','umi']
            df = pd.concat([df,tmp])
    df = df.drop_duplicates(['snv','barcode','umi'])
    df['num'] = 1
    pi = pd.pivot_table(df,index=['barcode','snv'],values='num',aggfunc='count').reset_index()
    pi['num'] = 1
    deep = pd.pivot_table(df,index=['barcode','snv'],values='num',aggfunc='count').reset_index()# how many read suppose this snv
#     deep_name = os.path.join(out_dir,"SNV.depth")
#     stat_name = os.path.join(out_dir,"SNV.stat")
#     pi.to_csv(stat_name,index=None)
#     deep.to_csv(deep_name,index=None)
    if x_offset != None and bin_size != 1:
        pi['x'] = pi['barcode'].map(lambda x : (int(x.split('_')[0]) - x_offset)//bin_size*bin_size)
        pi['y'] = pi['barcode'].map(lambda x : (int(x.split('_')[1]) - y_offset)//bin_size*bin_size)
        pi['location'] = pi['x'].map(str)+'_'+pi['y'].map(str)
        pi = pd.pivot_table(pi,index=['location','snv'],values='num',aggfunc='sum')
        pi = pi.reset_index()
        pi = pi.rename({'location':'barcode'},axis=1)
    else:
        pi = pd.pivot_table(pi,index=['barcode','snv'],values='num',aggfunc='sum')
        pi = pi.reset_index()

    if x_offset != None and bin_size != 1:
        deep['x'] = deep['barcode'].map(lambda x : (int(x.split('_')[0]) - x_offset)//bin_size*bin_size)
        deep['y'] = deep['barcode'].map(lambda x : (int(x.split('_')[1]) - y_offset)//bin_size*bin_size)
        deep['location'] = deep['x'].map(str)+'_'+deep['y'].map(str)
        deep = pd.pivot_table(deep,index=['location','snv'],values='num',aggfunc='sum')
        deep = deep.reset_index()
        deep = deep.rename({'location':'barcode'},axis=1)
    else:
        deep = pd.pivot_table(pi,index=['barcode','snv'],values='num',aggfunc='sum')
        deep = deep.reset_index()
    
    rawsample_matrix_path = sample_matrix_path
    sample_matrix_path = os.path.join(rawsample_matrix_path,f'{sample}_alt')
    get_mtx(pi,sample_matrix_path)

    sample_matrix_path_depth = os.path.join(rawsample_matrix_path,f'{sample}_depth')
    get_mtx(deep,sample_matrix_path_depth)

    ref_vcf_list = glob.glob(f"{sample_tmp_path}/*.ref.stat")
    df = pd.DataFrame(columns=['snv','barcode','umi'])
    for i in ref_vcf_list:
        if os.path.getsize(i):
            tmp=pd.read_csv(i,header=None)
            tmp.columns=['snv','barcode','umi']
            df = pd.concat([df,tmp])
    df = df.drop_duplicates(['snv','barcode','umi'])
    df['num'] = 1
    pi = pd.pivot_table(df,index=['barcode','snv'],values='num',aggfunc='count').reset_index()
    pi['num'] = 1
    if x_offset != None and bin_size != 1:
        pi['x'] = pi['barcode'].map(lambda x : (int(x.split('_')[0]) - x_offset)//bin_size*bin_size)
        pi['y'] = pi['barcode'].map(lambda x : (int(x.split('_')[1]) - y_offset)//bin_size*bin_size)
        pi['location'] = pi['x'].map(str)+'_'+pi['y'].map(str)
        pi = pd.pivot_table(pi,index=['location','snv'],values='num',aggfunc='sum')
        pi = pi.reset_index()
        pi = pi.rename({'location':'barcode'},axis=1)
    else:
        pi = pd.pivot_table(pi,index=['barcode','snv'],values='num',aggfunc='sum')
        pi = pi.reset_index()
    
    sample_matrix_path_ref = os.path.join(rawsample_matrix_path,f'{sample}_ref')
    get_mtx(pi,sample_matrix_path_ref)


    if removetmp:
        for i in (alt_vcf_list + ref_vcf_list):
            os.remove(i)


def get_mtx(pi,outdir_mtx):
    if not os.path.exists(outdir_mtx):
        os.makedirs(outdir_mtx)
    barcodes = pi['barcode'].drop_duplicates(keep='first').reset_index()
    barcodes_dict = dict(zip(barcodes['barcode'], barcodes.index + 1))
    features = pi['snv'].drop_duplicates(keep='first').reset_index()
    features_dict = dict(zip(features['snv'], features.index + 1))
    pi['barcode'] = pi['barcode'].apply(lambda x: barcodes_dict[x])
    pi['snv'] = pi['snv'].apply(lambda x: features_dict[x])
    hd = pd.DataFrame([['%%MatrixMarket matrix coordinate integer general'],\
                                   ['%'], \
                                   [' '.join([str(len(features)), str(len(barcodes)), str(len(pi))])]])
    hd.to_csv(outdir_mtx+'/matrix.mtx.gz', compression='gzip', sep = '\t', index=False, header=False)
    pi[['snv', 'barcode', 'num']].to_csv(outdir_mtx+'/matrix.mtx.gz', compression='gzip', mode='a+',\
                                                                                                 sep=' ', header=False, index=False)
    barcodes['barcode'].to_csv(outdir_mtx+'/barcodes.tsv.gz', compression='gzip', sep='\t', header=False, index=False)
    features['snv_id'] = features['snv']
    features['feature_type'] = 'Gene Expression'
    features[['snv', 'snv_id', 'feature_type']].to_csv(outdir_mtx+'/features.tsv.gz', compression='gzip', sep='\t', header=False, index=False)


import pysam
import os 
import shutil
import pandas as pd
import numpy as np

from .utils import logger_config,DirCheck,RunCMD
##########################################
####  BarcodeUMIBinding Part 
##########################################
def Code_barcode(barcode):
    code = {1:'AA',2:'AT',3:'AC',4:'AG',5:'TA',6:'TT',7:'TC',8:'TG',9:'CA',0:'CC'}
    barcode_list = list()
    for m in barcode:
        barcode_list.append(code[int(m)])
    return ''.join(barcode_list)

def Decoding_umi(umi):
    umi_dict = {0:'A',1:'C',2:'G',3:'T'}
    umi_num = np.base_repr(int(umi,16),base=4)
    umi_list = list()
    for i in umi_num:
        umi_list.append(umi_dict[int(i)])
    return ''.join(umi_list)

def LoadLocation(gem,xoff,yoff):
    is_gzip = gem.endswith('.gz')
    if is_gzip:
        df = pd.read_csv(gem,sep='\t',compression='gzip', comment='#')
    else:
        df = pd.read_csv(gem,sep='\t', comment='#')
    df = df.drop_duplicates(['x','y'])
    df['bam_x'] = df['x'] + xoff
    df['bam_y'] = df['y'] + yoff
    df['location'] = df['bam_x'].map(str)+'-'+df['bam_y'].map(str)
    uniq_coo = pd.DataFrame(df['location'])
    uniq_coo['num'] = 1
    location_dict = dict(zip(uniq_coo['location'],uniq_coo['num']))
    return location_dict

def BarcodeUMIBinding(bam,outbam,stereo = False,barcode_tag = "CR",umi_tag = 'UR',location_dict = None):

    inputfile = pysam.AlignmentFile(bam, "rb")
    output = pysam.AlignmentFile(outbam, 'wb', template = inputfile)

    if not stereo:
        for read in inputfile:
            _barcode = read.has_tag(barcode_tag)
            _umi = read.has_tag(umi_tag)
            if _barcode and _umi:
                barcode = read.get_tag(barcode_tag)
                umi = read.get_tag(umi_tag)
                BI = barcode + "-" + umi
                read.set_tag('LY', BI)
                output.write(read)
    else:
        for read in inputfile:
            x = str(read.get_tag('Cx'))
            y = str(read.get_tag('Cy'))
            umi = str(read.get_tag('UR'))

            if location_dict != None:
                location = x + '-' + y
                intissue = location_dict.get(location,False)
                if intissue:
                    new_barcode = Code_barcode(x) + '-' + Code_barcode(y)
                    new_umi = Decoding_umi(umi)
                    encode_barcode = new_barcode + '-' + new_umi
                    read.set_tag('LY', encode_barcode)
                    read.set_tag('CB', new_barcode)
                    read.set_tag('UM', new_umi)
                    output.write(read)
    output.close()
    inputfile.close()

##########################################
####  GATK process Part
##########################################
def GATKprocess(config,outdir,tmpdir,proxy,fasta,dbsnp):
    logger = logger_config("GATKprocess")
    Markduplicate_cmd = config.picard_cmd("MarkDuplicates", tmpdir)
    Markduplicate_cmd.extend(
            [   
                "-I",
                f"{tmpdir}/{proxy}.tag.bam",
                "-O",
                f"{tmpdir}/{proxy}.dup.bam",
                "-M",
                f"{tmpdir}/{proxy}.txt",
                "--TMP_DIR",
                f"{tmpdir}",
                "--REMOVE_SEQUENCING_DUPLICATES",
                "true",
                "--BARCODE_TAG",
                "LY",
            ]
        )
    RunCMD(Markduplicate_cmd, "MarkDuplicates",f"{proxy}",logger)

    Addsm_cmd = config.picard_cmd("AddOrReplaceReadGroups", tmpdir)
    Addsm_cmd.extend(
        [
            "-I",
            f"{tmpdir}/{proxy}.dup.bam",
            "-SM",
            "tumor",
            "-PL",
            "iDrop",
            "-LB",
            "lib1",
            "-PU",
            "unit1",
            "-ID",
            "3",
            "-O",
            f"{tmpdir}/{proxy}.sm.bam",
        ]
    )
    RunCMD(Addsm_cmd,"AddOrReplaceReadGroups",f"{proxy}",logger)

    SplitNCigarReads_cmd = config.gatk_cmd("SplitNCigarReads",tmpdir)
    SplitNCigarReads_cmd.extend(
        [
            "-R",
            fasta,
            "-I",
            f"{tmpdir}/{proxy}.sm.bam",
            "-O",
            f"{tmpdir}/{proxy}.nc.bam",
            "--max-reads-in-memory",
            300000,
        ]
    )
    RunCMD(SplitNCigarReads_cmd,"SplitNCigarReads",f"{proxy}",logger)

    BaseRecalibrator_cmd = config.gatk_cmd("BaseRecalibrator")
    BaseRecalibrator_cmd.extend(
        [
            "-R",
            fasta,
            "-I",
            f"{tmpdir}/{proxy}.nc.bam",
            "-O",
            f"{tmpdir}/{proxy}.base",
            "--use-original-qualities",
            "--known-sites",
            dbsnp,
        ]
    )
    RunCMD(BaseRecalibrator_cmd,"BaseRecalibrator",f"{proxy}",logger)

    ApplyBQSR_cmd = config.gatk_cmd("ApplyBQSR")
    ApplyBQSR_cmd.extend(
        [
            "-R",
            fasta,
            "-I",
            f"{tmpdir}/{proxy}.nc.bam",
            "-O",
            f"{outdir}/{proxy}.rdfcall.bam",
            "-bqsr",
            f"{tmpdir}/{proxy}.base",
            "--add-output-sam-program-record",
            "--use-original-qualities",
        ]
    )
    RunCMD(ApplyBQSR_cmd,"ApplyBQSR",f"{proxy}",logger)


def PerpareBAMforCalling_process(bam,outdir,proxy,barcode_tag,umi_tag,stereo,gem,xsetoff,ysetoff,config,fasta,dbsnp,tmpdir,removetmp):


    logger = logger_config("PerpareBAMforCalling")
    if tmpdir == None:
        tmpdir = os.path.join(outdir,"tmp")
    
    logger.info(f"Tmp dir: {tmpdir}")
    DirCheck(tmpdir,create = True)
    
    if stereo:
        LocationDict = LoadLocation(gem,xsetoff,ysetoff)
    else:
        LocationDict = None

    TagBAM = f"{tmpdir}/{proxy}.tag.bam"
    BarcodeUMIBinding(
        bam = bam,
        outbam = TagBAM,
        stereo = stereo,
        barcode_tag = barcode_tag,
        umi_tag = umi_tag,
        location_dict = LocationDict
        )
    logger.info(f"BarcodeUMIBinding: Finish")

    GATKprocess(
        config = config,
        outdir = outdir,
        tmpdir = tmpdir,
        proxy = proxy,
        fasta = fasta,
        dbsnp = dbsnp
        )

    if removetmp:
        logger.info(f"Remove Tmp Dir: {tmpdir}")
        shutil.rmtree(tmpdir)



import numpy as np
import pandas as pd
import scanpy as sc
sc.settings.verbosity = 3
import matplotlib.pyplot as plt
import seaborn as sns

import subprocess
import re
import os 
import sys
import gc
from multiprocessing import Pool, cpu_count

import sklearn.metrics, sklearn.cluster
from sklearn.decomposition import PCA
from sklearn.decomposition import TruncatedSVD
from sklearn.neighbors import kneighbors_graph, radius_neighbors_graph

from time import ctime
import scipy.sparse as sps
import igraph as ig
import leidenalg as la
import sklearn

import sklearn.metrics, sklearn.cluster
from sklearn.decomposition import PCA
from sklearn.decomposition import TruncatedSVD
from sklearn.neighbors import kneighbors_graph, radius_neighbors_graph
import sklearn.neighbors
from tqdm import tqdm
# from scanpy import _utils as utils



def rank_weight(csr_mtx, rank_cutoff, transpose=False, ret='value', include_self=True, alpha=1, decay='distance2'):
    from scipy.stats import rankdata
    if transpose:
        tmp = csr_mtx.tocoo().T.tocsr()
    else:
        tmp = csr_mtx.tocoo().tocsr()

    data = tmp.data[tmp.data != 0]
    ptr = tmp.indptr
    row, col = np.nonzero(tmp)
    
    value = np.array([])
    data2 = np.array([])
    row2 = np.array([])
    col2 = np.array([])
    for i in range(tmp.shape[0]):
        row_data = data[ptr[i]:ptr[i+1]].copy()
        row_ind = row[ptr[i]:ptr[i+1]]
        col_ind = col[ptr[i]:ptr[i+1]]
        
        if not include_self:
            indices = row_ind != col_ind
            row_data = row_data[indices]
            row_ind = row_ind[indices]
            col_ind = col_ind[indices]

        rank_data = rankdata(- row_data, method='min')

        indices = rank_data <= rank_cutoff

        
        value = np.append(value, row_data[indices])
        data2 = np.append(data2, rank_data[indices])
        col2 = np.append(col2, col_ind[indices])
        row2 = np.append(row2, row_ind[indices])
        
    if ret == 'rank':
        return sps.csr_matrix((data2+1, (row2, col2)),shape=tmp.shape)
    if ret == 'value':
        return sps.csr_matrix((value, (row2, col2)),shape=tmp.shape)
    if ret == 'weight':
        w = sps.csr_matrix((value, (row2, col2)),shape=tmp.shape)
        w = dist_decay(w, kernel=decay, alpha=alpha)
        # w = weight_norm(w, alpha=alpha)
        return w

def weight_norm(csr_mtx, alpha): ### normalize the total weights per spot to 1 
    con = csr_mtx.copy()
    con.data = con.data ** alpha
    scalers = np.array(con.sum(axis=1))[:, 0]
    scalers = 1 / scalers
    sklearn.utils.sparsefuncs.inplace_row_scale(con, scalers)
    return con
        
def dist_decay(con, kernel='Gaussion', alpha=2):
    
    r = con.data.mean()/alpha
    
    if kernel == 'Gaussian':
        con.data = np.exp(-con.data**2/(2*r**2))
    
    if kernel == 'Exponential':
        con.data = np.exp(-con.data/(2*r**2))
    
    if kernel == 'Laplacian':
        con.data = np.exp(-con.data/(2*r))
    
    if kernel == 'Cauchy':
        con.data = np.exp(1/(con.data**2/r + 1))
        
    if kernel == 'distance':
        con.data = 1/(1+con.data)
    
    if kernel == 'distance2':
        con.data = 1/(1+con.data**2)
    return con

def build_connect(adata, radius, include_self=False,scaling=True, norm=True, decay='Gaussian', alpha=2, rank_cutoff=100, rank2=True, n_neighbors=15, method='radius', use_raw=False):
    
    if method == 'radius':
        con  = radius_neighbors_graph(adata.obsm['spatial'], radius=radius, include_self=include_self, mode='distance')
               
    
    if method == 'knn':
        con = kneighbors_graph(adata.obsm['spatial'], n_neighbors=n_neighbors, include_self=include_self, mode='distance')
    
    con = dist_decay(con, kernel=decay, alpha=alpha)
    
    if norm:   ### normalize the total weights per spot to 1 
        con = weight_norm(con, alpha=1)           
    print('Step1: cell-cell connectivity calculation finished,', ctime()) 
    
    if use_raw:
        adata = adata.raw.to_adata()

    reads = adata.X.T.copy()
    
    v = mul_rank_weight(reads, rank_cutoff=rank_cutoff, ret='rank', alpha=1, transpose=False, nthreads=10)
   
    v = reads.multiply(v > 0)
    v_min = mul_rank_weight(-v, rank_cutoff=1, ret='value', alpha=1, transpose=False, nthreads=10)
    v_min = (-v_min).max(axis=1).toarray()[:, 0]
    sklearn.utils.sparsefuncs.inplace_row_scale(v, 1/v_min)

    print('Step2: ranking SNV expression values finished,', ctime())

    s_con = mul_sparse_dot(con, v.T, nthreads=20)
    
    counts = (con != 0).astype('float64')
    value = (v.T != 0).astype('float64')

    mode = np.array(counts.sum(axis=1))[:, 0]
    p = np.array(value.sum(axis=0))[0, :]/value.shape[0]

    scaler_mode = 1/mode
    scaler_p = 1/p
    
    counts = counts.dot(value)
    # return counts, value, scaler_mode, scaler_p
    sklearn.utils.sparsefuncs.inplace_row_scale(counts.T, scaler_p)
    sklearn.utils.sparsefuncs.inplace_row_scale(counts, scaler_mode)
    counts.data = np.log10(1 + counts.data)
    
    s_con = s_con.multiply(counts)

    print('Step3: cell-SNV connectivity calculation finished,', ctime())

    if rank2:
        s_con = mul_rank_weight(s_con, rank_cutoff=100, ret='value', nthreads=20)
        print('Step4: ranking cell-SNV connectivity finished,', ctime())

    ss_con = mul_sparse_dot(v, s_con, nthreads=10)
    print('Step5: SNV-SNV connectivity calculation finished,', ctime())
    
    return ss_con





def create_graph(con, directed=True):
    import igraph as ig
    sources, targets = con.nonzero()
    weights = con[sources, targets]
    if isinstance(weights, np.matrix):
        weights = weights.A1
        
    g = ig.Graph(directed=directed)
    g.add_vertices(con.shape[0]) 
    g.add_edges(list(zip(sources, targets)))
    g.es['weight'] = weights
    
    return g


def partition_leiden(time, g, ptype, max_comm_size,snv_var):
    part = la.find_partition(g, partition_type=ptype, seed=123, resolution_parameter=time, max_comm_size=max_comm_size, weights=np.array(g.es['weight']).astype(np.float64))
    y_predict = np.array(part.membership)
    peak_labels_df = pd.DataFrame(y_predict, index = snv_var, columns=['group']) 
    accum = np.sum(peak_labels_df.value_counts() > 1) / len(peak_labels_df.value_counts())
    return accum



def build_sg(snv, con, save=None,resolution=None, rank_cutoff=50, ptype=la.RBConfigurationVertexPartition, max_comm_size=0, syn=False):
    con = mul_rank_weight(con, rank_cutoff=rank_cutoff, ret='value',transpose=False, include_self=False, nthreads=20)
    if syn:
        con = (con + con.T) * 0.5
        con = weight_norm(con, alpha=1)

        g = create_graph(con, directed=False)
        
    else:
        con = weight_norm(con, alpha=1)
        g = create_graph(con, directed=True)
        
    if resolution == None:
        snv_var = snv.var_names.copy()
        args = [(time, g, ptype, max_comm_size,snv_var) for time in range(1,10)]

        
        with Pool(processes = 2) as pool:
            accum_list = list(tqdm(pool.starmap(partition_leiden, args), total=len(range(1,10))))
    
        plt.figure(figsize=(5, 5))
        plt.plot(accum_list, marker='o', linestyle='--', color='black', markerfacecolor='black', markeredgecolor='black')
        plt.title("Proportion of Communities with Multiple Members using Leiden Algorithm")
        plt.xlabel("Resolution Parameter (Time)")
        plt.ylabel("Proportion of Multi-member Communities")
        plt.grid(False)
        plt.show()
    else:
        part = la.find_partition(g, partition_type=ptype,  seed=123, resolution_parameter=resolution, max_comm_size=max_comm_size, weights=np.array(g.es['weight']).astype(np.float64))
        y_predict = np.array(part.membership)
        
        peak_labels_df = pd.DataFrame(y_predict, index = snv.var_names, columns = ['group'])
        
        snv.var['snv_group'] = y_predict
        
        counts = snv.var['snv_group'].value_counts()
        
        groups = list(set(y_predict))
        sg_matrix = np.array([snv.X.T[np.where(y_predict==x)[0],:].sum(axis=0) for x in groups])
        sg_matrix = sg_matrix[:,0,:].T
        
        sg_df = pd.DataFrame(sg_matrix, index = snv.obs_names, columns = groups)
        snv.obsm['snv_group'] = sg_df.values
        peak_values = peak_labels_df.value_counts().reset_index()
        peak_values['group'] = 'sg' + peak_values['group'].map(str)
        peak_values = peak_values.set_index('group')
        peak_values.columns = ['snv_count']
        ac = sc.AnnData(snv.obsm['snv_group'].copy(), obs=snv.obs)
        ac.var_names = ['sg'+str(x) for x in ac.var_names]
        ac.var['snv_count'] = peak_values['snv_count']
        ac.obsm['spatial'] = snv.obsm['spatial']
        return snv,ac


def snp_norm(snp, adata, method='umi'):
    if method == 'umi':
        scalers = 10000/adata.obs['total_counts'].values
        sklearn.utils.sparsefuncs.inplace_row_scale(snp.X, scalers)
    if method == 'gene':
        scalers = 10000/adata.obs['n_genes_by_counts'].values
        sklearn.utils.sparsefuncs.inplace_row_scale(snp.X, scalers)
    if method == 'total':
        sc.pp.calculate_qc_metrics(snp, percent_top=None, log1p=False, inplace=True)
        scalers = np.array(10000/snp.X.sum(axis=0))[0]
        sklearn.utils.sparsefuncs.inplace_column_scale(snp.X, scalers)
    snp.X.data = np.log2(snp.X.data+1)
    

import threading

class sparse_dot(threading.Thread):
    def __init__(self, csr_a, b, count, ret_dict):
        threading.Thread.__init__(self)
        self.csr_a = csr_a
        self.b = b
        self.count = count
        self.ret_dict = ret_dict
    
    def run(self):
        ret = self.csr_a.dot(self.b)
        self.ret_dict[self.count] = ret
    
    # def __del__(self):
    #     print('thread %d is ending' % self.count)

class sparse_rank(threading.Thread):
    def __init__(self, csr_mtx, rank_cutoff, ret, include_self, count, ret_dict, alpha):
        threading.Thread.__init__(self)
        self.csr_mtx = csr_mtx
        self.rank_cutoff = rank_cutoff
        self.ret = ret
        self.include_self = include_self
        self.count = count
        self.ret_dict = ret_dict
        self.alpha = alpha
    
    def run(self):
        tmp = rank_weight(self.csr_mtx, 
                          rank_cutoff=self.rank_cutoff, 
                          transpose=False, 
                          ret=self.ret, 
                          include_self=self.include_self,
                          alpha=self.alpha)
        self.ret_dict[self.count] = tmp


def mul_rank_weight(csr_mtx, rank_cutoff, nthreads, transpose=False, ret='value', include_self=True, alpha=1):

    if transpose:
        csr_mtx = csr_mtx.tocoo().T.tocsr()
        
    nrow = csr_mtx.shape[0]

    step = nrow // nthreads
    if (step * nthreads) != nrow:
        step = nrow // nthreads + 1
    slices = list(range(0, nrow, step))
    slices.append(nrow)
    
    ret_dict = dict()
    threads = list()
    for i in range(nthreads):
        a = csr_mtx[slices[i]:slices[i+1], :]
        count = i
        thread = sparse_rank(a, 
                             rank_cutoff=rank_cutoff, 
                             ret=ret, 
                             include_self=include_self, 
                             count=count, 
                             ret_dict=ret_dict,
                             alpha=alpha)
        threads.append(thread)
        thread.start()
    
    for thread in threads:
        thread.join()
    
    return sps.vstack([ret_dict[i] for i in range(nthreads)])
        

def mul_sparse_dot(csr_a, b, nthreads):
    
    if not sps.isspmatrix_csr(csr_a):
        csr_a = csr_a.tocsr()
        
    nrow = csr_a.shape[0]

    step = nrow // nthreads
    if (step * nthreads) != nrow:
        step = nrow // nthreads + 1
    slices = list(range(0, nrow, step))
    slices.append(nrow)
    
    ret_dict = dict()
    threads = list()
    for i in range(nthreads):
        a = csr_a[slices[i]:slices[i+1], :]
        count = i
        thread = sparse_dot(a, b, count, ret_dict)
        threads.append(thread)
        thread.start()
    
    for thread in threads:
        thread.join()
    
    return sps.vstack([ret_dict[i] for i in range(nthreads)])
        

def sg_cluster(adata, res=0.5, npcs=15, neighbors=15, norm=False, subset=None):
    sc.settings.verbosity = 1
    
    ac = sc.AnnData(adata.obsm['snv_group'].copy(), obs=adata.obs)
    if subset != None:
        t = adata.var['snv_group'].value_counts()
        t = t[t>subset].index.tolist()
        t = list(map(str,t))
        ac = ac[:,ac.var_names.isin(t)]
    ac.var_names = ['sg'+str(x) for x in ac.var_names]
    ac.layers['raw_counts'] = ac.X.copy()
    if norm:
        sc.pp.normalize_total(ac, target_sum=1e4)
    sc.pp.log1p(ac)
    sc.pp.scale(ac, max_value=10)
    sc.pp.pca(ac, svd_solver='full')
    sc.pl.pca_variance_ratio(ac, n_pcs=50)
    sc.pp.neighbors(ac, n_neighbors=neighbors, n_pcs=npcs)
    sc.tl.umap(ac)
    sc.tl.leiden(ac, resolution=res)
    # sc.pl.umap(ac, color=['leiden'])
    adata.obs['leiden_sg'] = ac.obs['leiden']
    adata.obsm['X_umap'] = ac.obsm['X_umap']
    return ac

# process

def snv2avinput(
    snv,
    sample_name,
    annovar,
    annovar_ref,
    gtf,
    spe,
    ref_name,
    outdir,
    overwrite = False
):

    if overwrite:
        avi = pd.DataFrame(snv.var_names)
        avi.columns=['snv']
        avi['chrom'] = avi['snv'].apply(lambda x : x.split('_')[0])
        avi['site1'] = avi['snv'].apply(lambda x : x.split('_')[1].split(':')[0])
        avi['site2'] = avi['site1']
        avi['ref_base'] = avi['snv'].apply(lambda x : x.split('_')[1].split(':')[1].split('>')[0])
        avi['alt_base'] = avi['snv'].apply(lambda x : x.split('_')[1].split(':')[1].split('>')[1])
        avi = avi[['chrom','site1','site2','ref_base','alt_base']]
        if not os.path.exists(outdir):
            os.mkdir(outdir)
        avi.to_csv(f'{outdir}/{sample_name}.avinput',index=None,sep='\t',header=None)
        with open(f'{outdir}/{sample_name}.anno.sh','w') as f:
            f.write(f'perl {annovar} {outdir}/{sample_name}.avinput {annovar_ref} --out {outdir}/{sample_name} --buildver {spe} --remove --protocol {ref_name} --operation g --nastring .\n')
        subprocess.call(f"bash {outdir}/{sample_name}.anno.sh",shell=True)
    
    annotation = pd.read_csv(f'{outdir}/{sample_name}.{spe}_multianno.txt',sep='\t')
    annotation['snp'] = annotation['Chr']+'_'+annotation['Start'].map(str) + ':' + annotation['Ref'] + '>' +annotation['Alt']
    annotation = annotation.set_index('snp')
    gene_info = pd.read_csv(gtf, sep='\t', comment='#', usecols=[0,2,3,4, 8], names=['chr','type','start','end', 'info'])
    gene_info = gene_info[gene_info['type'] == 'gene']
    gene_info['gene_id'] = gene_info['info'].str.extract(r'gene_id "([^"]+)";')
    gene_info = gene_info.drop_duplicates(subset='gene_id').reset_index(drop=True)
    gene_info['gene_name'] = gene_info['info'].str.extract(r'gene_name "([^"]+)";')
    gene_info['gene_lenth'] = gene_info['end'] - gene_info['start']
    gene_annotation_dict = dict(zip(gene_info['gene_id'],gene_info['gene_name']))
    gene_len_dict = dict(zip(gene_info['gene_id'],gene_info['gene_lenth']))
    def map_gene(genecode):
        return gene_annotation_dict.get(genecode,'None')
    def map_len(genecode):
        return str(gene_len_dict.get(genecode,'None'))
    annotation['gene_name'] = annotation[f'Gene.{ref_name}'].apply(lambda x : ';'.join(list(map(map_gene,x.split(';')))))
    annotation['gene_lenth'] = annotation[f'Gene.{ref_name}'].apply(lambda x : ';'.join(list(map(map_len,x.split(';')))))
    annotation = annotation[[f'Func.{ref_name}',f'ExonicFunc.{ref_name}','gene_name','gene_lenth']]
    snv.var[f'Func'] = annotation[f'Func.{ref_name}']
    snv.var[f'ExonicFunc'] = annotation[f'ExonicFunc.{ref_name}']
    snv.var['gene_name'] = annotation['gene_name']
    snv.var['gene_lenth'] = annotation['gene_lenth']
    return snv

def summary(region):
    if re.match(r'^(ncRNA_)?exonic(;splicing)?$', region):
        part = "Exonic"
    elif re.match(r'^(ncRNA_)?intronic$', region) or re.match(r'^(ncRNA_)?splicing$', region):
        part = "Intronic"
    elif re.match(r'^UTR', region):
        part = "UTR"
    else:
        part = "Intergenic"
    return part

def processsnv(
    sample_name,
    snv,
    snv_depth,
    gtf,
    annovar_ref,
    annovar_spe,
    annovar_ref_name,
    annovar,
    outdir,
    thrshold = 20,
    min_cells = 5,
):
    snv.var['SNVDepth'] = np.sum(snv_depth.X,axis = 0).reshape(-1,1)
    snv.obs['TotalDepth'] = np.sum(snv_depth.X,axis = 1)
    snv.var['SNVCount'] = np.sum(snv.X,axis = 0).reshape(-1,1)
    sc.pp.filter_genes(snv,min_cells = 1)
    sc.pp.filter_genes(snv_depth,min_cells = 1)
    thrshold = 20
    sub_snv = snv[:,snv.var['SNVDepth'] >= thrshold].copy()
    sub_snv_depth = snv_depth[:,snv.var['SNVDepth'] >= thrshold].copy()
    sc.pp.filter_genes(sub_snv,min_cells = min_cells)
    sub_snv = snv2avinput(
        sub_snv,
        sample_name,
        annovar = annovar,
        annovar_ref = annovar_ref,
        gtf = gtf,
        spe = annovar_spe,
        ref_name = annovar_ref_name,
        outdir = outdir,
        overwrite=True
    )
    sub_snv.var['Func_L0'] = sub_snv.var['Func'].apply(lambda x : summary(x))
    return sub_snv

def _normalize_data(X, counts, after= None):
    from sklearn.utils.sparsefuncs import inplace_row_scale
    X = X.copy()
    after = np.median(counts_greater_than_zero, axis=0) if after is None else after
    counts += counts == 0
    counts = counts / after
    inplace_row_scale(X, 1 / counts)
    return X

def normalize_with_rna(snv,rna,target_sum = 1e4):
    snv.obs['UMI_counts'] = rna.obs['total_counts']
    umi_per_cell = np.ravel(snv.obs['UMI_counts'])
    snv.layers['norm'] = np.log1p(_normalize_data(snv.X,umi_per_cell,after = target_sum))
    snv.obs['snvperumi'] = np.sum(snv.layers['norm'],axis = 1)
    matrix = snv.layers['norm'].todense()
    matrix = np.where(matrix>0,1,0)
    snv.obs['SNVtypes'] = np.sum(matrix,axis = 1 ).reshape(-1,1)

from anndata import AnnData
from scipy.sparse import csc_matrix
def bulid_windows(snv,window_size = 100000,basis = 'basis'):
    snv_df = snv.to_df().T
    snv_df['chrom'] = snv_df.index.map(lambda x: x.split('_')[0])
    snv_df['num'] = snv_df.index.map(lambda x: str(int(x.split('_')[1].split(':')[0])//window_size))
    snv_df['windows'] = snv_df['chrom'] +'@'+ snv_df['num']
    del snv_df['chrom']
    del snv_df['num']
    snv_df = snv_df.groupby('windows').sum().T
    snv_gene_adata = AnnData(snv_df)
    snv_gene_adata.X = csc_matrix(snv_gene_adata.X)
    snv_gene_adata.obsm[basis] = snv.obsm[basis]
    return snv_gene_adata


def get_min_distance(adata,basis = 'spatial'):
    locations = adata.obsm[basis].copy()
    distances = np.linalg.norm(locations[:, np.newaxis] - locations, axis=2)
    np.fill_diagonal(distances, np.inf)
    min_distances = np.min(distances, axis=1)
    average_min_distance = np.mean(min_distances)
    return average_min_distance

import igraph as ig
def netplot(snv_gene_adata,snv_group,ss_con,topn = 5,rank_cutoff=50,ret='value',transpose=False, include_self=False, nthreads=20,save = None):
    procon = mul_rank_weight(ss_con, rank_cutoff = rank_cutoff, ret=ret,transpose=transpose, include_self=include_self, nthreads=nthreads)
    procon = weight_norm(procon, alpha=1)
    inlist = snv_gene_adata[:,snv_gene_adata.var['snv_group'] == snv_group].var_names

    indices = [np.where(snv_gene_adata.var_names == gene)[0][0] for gene in inlist if gene in snv_gene_adata.var_names]
    sub_con = procon[indices, :][:, indices]
    g = create_graph(sub_con, directed=True)
    g.vs['name'] = inlist
    visual_style = {}
    visual_style["bbox"] = (300, 300) 
    visual_style["margin"] = 20
    visual_style["edge_color"] = "gray"
    degrees = g.degree()
    n = len(g.vs)
    top_indices = sorted(range(n), key=lambda i: degrees[i], reverse=True)[:topn]
    top_degrees = [degrees[i] for i in top_indices]
    max_degree_indices = sorted(range(len(degrees)), key=lambda i: degrees[i], reverse=True)[:topn]
    labels = ["" for _ in range(len(g.vs))]
    for idx in max_degree_indices:
        labels[idx] = g.vs[idx]["name"]
        
    min_top_degree, max_top_degree = min(top_degrees), max(top_degrees)
    node_sizes = [topn] * n 
    for idx in top_indices:
        if max_top_degree > min_top_degree: 
            size = ((degrees[idx] - min_top_degree) / (max_top_degree - min_top_degree) * 15) + 10  
        else:
            size = 25
        node_sizes[idx] = max(size, 5)
        
    layout = g.layout("mds")   #"grid"  "graphopt" "davidson_harel"  "drl" "mds" "kk"
    visual_style = {}
    visual_style["bbox"] = (350, 350)
    visual_style["edge_color"] = "gray"
    visual_style['edge_arrow_size'] = 0.1  
    if save != None:
        target = save
    
    ig.plot(g, layout=layout,
            vertex_size=node_sizes,edge_width = 0.2,
            vertex_color=['#E84B50' if idx in top_indices else '#00A1DF' for idx in range(n)],
            vertex_label=labels,
            **visual_style,target=save)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="spatialsnv",  
    version="1.0.2", 
    description="A toolkit for spatial SNV analysis", 
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Yi Liu",
    author_email="liuyi6@genomics.cn", 
    url="https://https://github.com/ly4dsg/SpatialSNV",  
    license="MIT", 
    packages=find_packages(), 
    py_modules=["spatialSNV"],
    python_requires=">=3.7",  
    install_requires=[
        "multiprocess==0.70.17",
        "click==8.1.8",
        "pysam==0.22.1",
        "pandas==1.5.3",
        "numpy==1.24.4",
        "scikit-learn==1.4.1.post1",
        "scipy==1.11.4",
        "igraph==0.11.5",
        "seaborn==0.13.2",
        "matplotlib==3.7.5",
        "scanpy==1.10.0",
        "tqdm",
        "cairocffi",
        "pycairo",
        "leidenalg",
    ],
    entry_points={
        "console_scripts": [
            "spatialsnvtools=spatialsnvtools.cli:main",  
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
)
